type phrase =
  | Def of string * Syntax.t
  | Infer of Syntax.t * Syntax.t
  | Check of Syntax.t * Syntax.t * Syntax.t
  | Run of Syntax.t * Syntax.t

type prog = phrase list
    
let register_defs prog =
  List.iter (function
	       | Def (x,re) -> Syntax.register x re
	       | Run _ | Infer _ | Check _ -> ()
	    ) prog

let print_run e ppf v =
  let v' = Run.run e v in
  Format.fprintf Format.std_formatter
    "In:%a@\nOut:%a@." 
    Syntax.print v
    Syntax.print v'

let run prog =
  register_defs prog;
  List.iter 
    (function
       | Infer (e,re) ->
	   Expr.check e;
	   Regexp.check re;
	   let q = Regexp.compile re in
	   let q' = Typecheck.infer e q in
	   if State.is_empty q' then
	     Format.fprintf Format.std_formatter "Empty !@."
	   else
	     Format.fprintf Format.std_formatter
	       "Valid input type:%a@." 
	       Syntax.print (Regexp.decompile q')
(*	       State.print_def q' *)
       | Check (e,rin,rout) ->
	   Expr.check e;
	   Regexp.check rin;
	   Regexp.check rout;
	   
	   let qin = Regexp.compile rin in
	   let qout = Regexp.compile rout in

	   let qin_inf = Typecheck.infer e qout in
	   (match Subtype.sample qin [qin_inf] with
	      | None -> Format.fprintf Format.std_formatter "Ok !@."
	      | Some ex ->
		  Format.fprintf Format.std_formatter
		    "Counter-example:@\n%a" 
		    (print_run e) ex)
       | Run (e,v) ->
	   Expr.check e;
	   Tree.check v;
	   print_run e Format.std_formatter v
       | _ -> ()
    ) prog
    
