open Tags
open Syntax

let seq v1 v2 = match v1,v2 with 
  | Eps,v | v,Eps -> v
  | v1,v2 -> Seq (v1,v2)

let rec run (v : (Tags.Symbol.t * Syntax.t) list)  = function
  | Elem (tag,e) -> 
      if SymbolSet.is_singleton tag 
      then Elem (tag, run v e)
      else
	(match v with 
	   | [] -> Eps 
	   | (tag, _)::_ -> Elem (SymbolSet.finite [tag], run v e))
  | Seq (e1,e2) -> seq (run v e1) (run v e2)
  | Eps -> Eps
  | Test (t, e1, e2) ->
      let q = Regexp.compile t in
      run v (if State.match_forest q v then e1 else e2)
  | Sub e -> (match v with [] -> Eps | (_, sub)::_ -> run (Tree.get sub) e)
  | Next e -> (match v with [] -> Eps | _::next -> run next e)
  | Name s -> run v (def s)
  | Star e as star -> run v (Seq (e, Next star))
  | _ -> assert false

let run e t = run (Tree.get t) e
