module Symbol = struct
  type t = string
  let compare = compare
end

module SSet = Slist.Make(Symbol)

module SymbolSet = struct
  type t =
    | Finite of SSet.t
    | Cofinite of SSet.t

  let equal = (=)
  let hash = Hashtbl.hash
  let compare = Pervasives.compare

  let is_empty = function
    | Finite s -> SSet.is_empty s
    | Cofinite _ -> false

  let inter s1 s2 = match s1,s2 with
    | Finite s1, Finite s2 -> Finite (SSet.inter s1 s2)
    | Finite s1, Cofinite s2 
    | Cofinite s2, Finite s1 -> Finite (SSet.diff s1 s2)
    | Cofinite s1, Cofinite s2 -> Cofinite (SSet.union s1 s2)

  let disjoint s1 s2 = match s1,s2 with
    | Finite s1, Finite s2 -> SSet.disjoint s1 s2
    | Finite s1, Cofinite s2 
    | Cofinite s2, Finite s1 -> SSet.subset s1 s2
    | Cofinite s1, Cofinite s2 -> false


  let union s1 s2 = match s1,s2 with
    | Finite s1, Finite s2 -> Finite (SSet.union s1 s2)
    | Finite s1, Cofinite s2 
    | Cofinite s2, Finite s1 -> Cofinite (SSet.diff s2 s1)
    | Cofinite s1, Cofinite s2 -> Cofinite (SSet.inter s1 s2)

  let diff s1 s2 = match s1,s2 with
    | Finite s1, Cofinite s2 -> Finite (SSet.inter s1 s2)
    | Finite s1, Finite s2 
    | Cofinite s2, Cofinite s1 -> Finite (SSet.diff s1 s2)
    | Cofinite s1, Finite s2 -> Cofinite (SSet.union s1 s2)

  let subset s1 s2 = is_empty (diff s1 s2)

  let any = Cofinite SSet.empty
  let empty = Finite SSet.empty

  let finite l = Finite (List.fold_right SSet.add l SSet.empty)
  let cofinite l = Cofinite (List.fold_right SSet.add l SSet.empty)

  let has sset s = match sset with
    | Finite sset -> SSet.mem s sset
    | Cofinite sset -> not (SSet.mem s sset)

  let print_sset ppf sset =
    let first = ref true in
    SSet.iter
      (fun x ->
	 if !first then first := false
	 else Format.fprintf ppf "|";
	 Format.fprintf ppf "%s" x
      )
      sset;
    if !first then Format.fprintf ppf "Empty"

  let print ppf = function
    | Finite sset -> 
	(match SSet.cardinal sset with
	   | 0 -> Format.fprintf ppf "^_"
	   | 1 -> Format.fprintf ppf "%s" (SSet.min_elt sset)
	   | _ -> Format.fprintf ppf "{%a}" print_sset sset)
    | Cofinite sset -> 
	(match SSet.cardinal sset with
	   | 0 -> Format.fprintf ppf "_"
	   | 1 -> Format.fprintf ppf "^%s" (SSet.min_elt sset)
	   | _ -> Format.fprintf ppf "{^%a}" print_sset sset)

  let sample = function
    | Finite sset -> SSet.min_elt sset
    | Cofinite _ -> "*"

  let is_singleton = function
    | Finite [_] -> true
    | _ -> false

  let is_any = function
    | Cofinite [] -> true
    | _ -> false
end


